# !/usr/bin/env python
# coding: utf-8

import urllib.parse

import requests
import requests.adapters
import logging

from requests.auth import HTTPBasicAuth

from legendary.models.exceptions import InvalidCredentialsError
from legendary.models.gql import *


class EPCAPI:
    _user_agent = 'UELauncher/11.0.1-14907503+++Portal+Release-Live Windows/10.0.19041.1.256.64bit'
    _store_user_agent = 'EpicGamesLauncher/14.0.8-22004686+++Portal+Release-Live'
    # required for the oauth request
    _user_basic = '34a02cf8f4414e29b15921876da36f9a'
    _pw_basic = 'daafbccc737745039dffe53d94fc76cf'
    _label = 'Live-EternalKnight'

    _oauth_host = 'account-public-service-prod03.ol.epicgames.com'
    _launcher_host = 'launcher-public-service-prod06.ol.epicgames.com'
    _entitlements_host = 'entitlement-public-service-prod08.ol.epicgames.com'
    _catalog_host = 'catalog-public-service-prod06.ol.epicgames.com'
    _ecommerce_host = 'ecommerceintegration-public-service-ecomprod02.ol.epicgames.com'
    _datastorage_host = 'datastorage-public-service-liveegs.live.use1a.on.epicgames.com'
    _library_host = 'library-service.live.use1a.on.epicgames.com'
    # Using the actual store host with a user-agent newer than 14.0.8 leads to a CF verification page,
    # but the dedicated graphql host works fine.
    # _store_gql_host = 'launcher.store.epicgames.com'
    _store_gql_host = 'graphql.epicgames.com'
    _artifact_service_host = 'artifact-public-service-prod.beee.live.use1a.on.epicgames.com'

    def __init__(self, lc='en', cc='US', timeout=10.0):
        self.log = logging.getLogger('EPCAPI')

        self.session = requests.session()
        self.session.headers['User-Agent'] = self._user_agent
        # increase maximum pool size for multithreaded metadata requests
        self.session.mount('https://', requests.adapters.HTTPAdapter(pool_maxsize=16))

        self.unauth_session = requests.session()
        self.unauth_session.headers['User-Agent'] = self._user_agent

        self._oauth_basic = HTTPBasicAuth(self._user_basic, self._pw_basic)

        self.access_token = None
        self.user = None

        self.language_code = lc
        self.country_code = cc

        self.request_timeout = timeout if timeout > 0 else None

    def get_auth_url(self):
        login_url = 'https://www.epicgames.com/id/login?redirectUrl='
        redirect_url = f'https://www.epicgames.com/id/api/redirect?clientId={self._user_basic}&responseType=code'
        return login_url + urllib.parse.quote(redirect_url)

    def update_egs_params(self, egs_params):
        # update user-agent
        if version := egs_params['version']:
            self._user_agent = f'UELauncher/{version} Windows/10.0.19041.1.256.64bit'
            self._store_user_agent = f'EpicGamesLauncher/{version}'
            self.session.headers['User-Agent'] = self._user_agent
            self.unauth_session.headers['User-Agent'] = self._user_agent
        # update label
        if label := egs_params['label']:
            self._label = label
        # update client credentials
        if 'client_id' in egs_params and 'client_secret' in egs_params:
            self._user_basic = egs_params['client_id']
            self._pw_basic = egs_params['client_secret']
            self._oauth_basic = HTTPBasicAuth(self._user_basic, self._pw_basic)

    def resume_session(self, session):
        self.session.headers['Authorization'] = f'bearer {session["access_token"]}'
        r = self.session.get(f'https://{self._oauth_host}/account/api/oauth/verify',
                             timeout=self.request_timeout)
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorMessage' in j:
            self.log.warning(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])

        # update other data
        session.update(j)
        self.user = session
        return self.user

    def start_session(self, refresh_token: str = None, exchange_token: str = None,
                      authorization_code: str = None, client_credentials: bool = False) -> dict:
        if refresh_token:
            params = dict(grant_type='refresh_token',
                          refresh_token=refresh_token,
                          token_type='eg1')
        elif exchange_token:
            params = dict(grant_type='exchange_code',
                          exchange_code=exchange_token,
                          token_type='eg1')
        elif authorization_code:
            params = dict(grant_type='authorization_code',
                          code=authorization_code,
                          token_type='eg1')
        elif client_credentials:
            params = dict(grant_type='client_credentials',
                          token_type='eg1')
        else:
            raise ValueError('At least one token type must be specified!')

        r = self.session.post(f'https://{self._oauth_host}/account/api/oauth/token',
                              data=params, auth=self._oauth_basic,
                              timeout=self.request_timeout)
        # Only raise HTTP exceptions on server errors
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorCode' in j:
            if j['errorCode'] == 'errors.com.epicgames.oauth.corrective_action_required':
                self.log.error(f'{j["errorMessage"]} ({j["correctiveAction"]}), '
                               f'open the following URL to take action: {j["continuationUrl"]}')
            else:
                self.log.error(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])
        elif r.status_code >= 400:
            self.log.error(f'EGS API responded with status {r.status_code} but no error in response: {j}')
            raise InvalidCredentialsError('Unknown error')

        self.session.headers['Authorization'] = f'bearer {j["access_token"]}'
        # only set user info when using non-anonymous login
        if not client_credentials:
            self.user = j

        return j

    def invalidate_session(self):  # unused
        _ = self.session.delete(f'https://{self._oauth_host}/account/api/oauth/sessions/kill/{self.access_token}',
                                timeout=self.request_timeout)

    def get_game_token(self):
        r = self.session.get(f'https://{self._oauth_host}/account/api/oauth/exchange',
                             timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_ownership_token(self, namespace, catalog_item_id):
        user_id = self.user.get('account_id')
        r = self.session.post(f'https://{self._ecommerce_host}/ecommerceintegration/api/public/'
                              f'platforms/EPIC/identities/{user_id}/ownershipToken',
                              data=dict(nsCatalogItemId=f'{namespace}:{catalog_item_id}'),
                              timeout=self.request_timeout)
        r.raise_for_status()
        return r.content

    def get_external_auths(self):
        user_id = self.user.get('account_id')
        r = self.session.get(f'https://{self._oauth_host}/account/api/public/account/{user_id}/externalAuths',
                             timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_game_assets(self, platform='Windows', label='Live'):
        r = self.session.get(f'https://{self._launcher_host}/launcher/api/public/assets/{platform}',
                             params=dict(label=label), timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_game_manifest(self, namespace, catalog_item_id, app_name, platform='Windows', label='Live'):
        r = self.session.get(f'https://{self._launcher_host}/launcher/api/public/assets/v2/platform'
                             f'/{platform}/namespace/{namespace}/catalogItem/{catalog_item_id}/app'
                             f'/{app_name}/label/{label}',
                             timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_launcher_manifests(self, platform='Windows', label=None):
        r = self.session.get(f'https://{self._launcher_host}/launcher/api/public/assets/v2/platform/'
                             f'{platform}/launcher', timeout=self.request_timeout,
                             params=dict(label=label if label else self._label))
        r.raise_for_status()
        return r.json()

    def get_user_entitlements(self):
        user_id = self.user.get('account_id')
        r = self.session.get(f'https://{self._entitlements_host}/entitlement/api/account/{user_id}/entitlements',
                             params=dict(start=0, count=5000), timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_game_info(self, namespace, catalog_item_id, timeout=None):
        r = self.session.get(f'https://{self._catalog_host}/catalog/api/shared/namespace/{namespace}/bulk/items',
                             params=dict(id=catalog_item_id, includeDLCDetails=True, includeMainGameDetails=True,
                                         country=self.country_code, locale=self.language_code),
                             timeout=timeout or self.request_timeout)
        r.raise_for_status()
        return r.json().get(catalog_item_id, None)

    def get_artifact_service_ticket(self, sandbox_id: str, artifact_id: str, label='Live', platform='Windows'):
        # Based on EOS Helper Windows service implementation. Only works with anonymous EOSH session.
        # sandbox_id is the same as the namespace, artifact_id is the same as the app name
        r = self.session.post(f'https://{self._artifact_service_host}/artifact-service/api/public/v1/dependency/'
                              f'sandbox/{sandbox_id}/artifact/{artifact_id}/ticket',
                              json=dict(label=label, expiresInSeconds=300, platform=platform),
                              params=dict(useSandboxAwareLabel='false'),
                              timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_game_manifest_by_ticket(self, artifact_id: str, signed_ticket: str, label='Live', platform='Windows'):
        # Based on EOS Helper Windows service implementation.
        r = self.session.post(f'https://{self._launcher_host}/launcher/api/public/assets/v2/'
                              f'by-ticket/app/{artifact_id}',
                              json=dict(platform=platform, label=label, signedTicket=signed_ticket),
                              timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_library_items(self, include_metadata=True):
        records = []
        r = self.session.get(f'https://{self._library_host}/library/api/public/items',
                             params=dict(includeMetadata=include_metadata),
                             timeout=self.request_timeout)
        r.raise_for_status()
        j = r.json()
        records.extend(j['records'])

        # Fetch remaining library entries as long as there is a cursor
        while cursor := j['responseMetadata'].get('nextCursor', None):
            r = self.session.get(f'https://{self._library_host}/library/api/public/items',
                                 params=dict(includeMetadata=include_metadata, cursor=cursor),
                                 timeout=self.request_timeout)
            r.raise_for_status()
            j = r.json()
            records.extend(j['records'])

        return records

    def get_user_cloud_saves(self, app_name='', manifests=False, filenames=None):
        if app_name:
            app_name += '/manifests/' if manifests else '/'

        user_id = self.user.get('account_id')

        if filenames:
            r = self.session.post(f'https://{self._datastorage_host}/api/v1/access/egstore/savesync/'
                                  f'{user_id}/{app_name}',
                                  json=dict(files=filenames),
                                  timeout=self.request_timeout)
        else:
            r = self.session.get(f'https://{self._datastorage_host}/api/v1/access/egstore/savesync/'
                                 f'{user_id}/{app_name}',
                                 timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def create_game_cloud_saves(self, app_name, filenames):
        return self.get_user_cloud_saves(app_name, filenames=filenames)

    def delete_game_cloud_save_file(self, path):
        url = f'https://{self._datastorage_host}/api/v1/data/egstore/{path}'
        r = self.session.delete(url, timeout=self.request_timeout)
        r.raise_for_status()

    def store_get_uplay_codes(self):
        user_id = self.user.get('account_id')
        r = self.session.post(f'https://{self._store_gql_host}/graphql',
                              headers={'user-agent': self._store_user_agent},
                              json=dict(query=uplay_codes_query,
                                        variables=dict(accountId=user_id)),
                              timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def store_claim_uplay_code(self, uplay_id, game_id):
        user_id = self.user.get('account_id')
        r = self.session.post(f'https://{self._store_gql_host}/graphql',
                              headers={'user-agent': self._store_user_agent},
                              json=dict(query=uplay_claim_query,
                                        variables=dict(accountId=user_id,
                                                       uplayAccountId=uplay_id,
                                                       gameId=game_id)),
                              timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def store_redeem_uplay_codes(self, uplay_id):
        user_id = self.user.get('account_id')
        r = self.session.post(f'https://{self._store_gql_host}/graphql',
                              headers={'user-agent': self._store_user_agent},
                              json=dict(query=uplay_redeem_query,
                                        variables=dict(accountId=user_id,
                                                       uplayAccountId=uplay_id)),
                              timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()



def get_boolean_choice(prompt, default=True):
    yn = 'Y/n' if default else 'y/N'

    choice = input(f'{prompt} [{yn}]: ')
    if not choice:
        return default
    elif choice[0].lower() == 'y':
        return True
    else:
        return False


def get_int_choice(prompt, default=None, min_choice=None, max_choice=None, return_on_invalid=False):
    if default is not None:
        prompt = f'{prompt} [{default}]: '
    else:
        prompt = f'{prompt}: '

    while True:
        try:
            if inp := input(prompt):
                choice = int(inp)
            else:
                return default
        except ValueError:
            if return_on_invalid:
                return None
            return_on_invalid = True
            continue
        else:
            if min_choice is not None and choice < min_choice:
                print(f'Number must be greater than {min_choice}')
                if return_on_invalid:
                    return None
                return_on_invalid = True
                continue
            if max_choice is not None and choice > max_choice:
                print(f'Number must be less than {max_choice}')
                if return_on_invalid:
                    return None
                return_on_invalid = True
                continue
            return choice


def sdl_prompt(sdl_data, title):
    tags = ['']
    if '__required' in sdl_data:
        tags.extend(sdl_data['__required']['tags'])

    print(f'You are about to install {title}, this application supports selective downloads.')
    print('The following optional packs are available (tag - name):')
    for tag, info in sdl_data.items():
        if tag == '__required':
            continue
        print(' *', tag, '-', info['name'])

    examples = ', '.join([g for g in sdl_data.keys() if g != '__required'][:2])
    print(f'Please enter tags of pack(s) to install (space/comma-separated, e.g. "{examples}")')
    print('Leave blank to use defaults (only required data will be downloaded).')
    choices = input('Additional packs [Enter to confirm]: ')
    if not choices:
        return tags

    for c in choices.strip('"').replace(',', ' ').split():
        c = c.strip()
        if c in sdl_data:
            tags.extend(sdl_data[c]['tags'])
        else:
            print('Invalid tag:', c)

    return tags


def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    Copied from python standard library as distutils.util.strtobool is deprecated.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return 1
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))




import logging
import plistlib
import os
import subprocess

_logger = logging.getLogger('CXHelpers')


def mac_get_crossover_version(app_path):
    try:
        plist = plistlib.load(open(os.path.join(app_path, 'Contents', 'Info.plist'), 'rb'))
        return plist['CFBundleShortVersionString']
    except Exception as e:
        _logger.debug(f'Failed to load plist for "{app_path}" with {e!r}')
        return None


def mac_find_crossover_apps():
    paths = ['/Applications/CrossOver.app']
    try:
        out = subprocess.check_output(['mdfind', 'kMDItemCFBundleIdentifier="com.codeweavers.CrossOver"'])
        paths.extend(out.decode('utf-8', 'replace').strip().split('\n'))
    except Exception as e:
        _logger.warning(f'Trying to find CrossOver installs via mdfind failed: {e!r}')

    valid = [p for p in paths if os.path.exists(os.path.join(p, 'Contents', 'Info.plist'))]
    found_tuples = set()

    for path in valid:
        version = mac_get_crossover_version(path)
        if not version:
            continue
        _logger.debug(f'Found Crossover {version} at "{path}"')
        found_tuples.add((version, path))

    return sorted(found_tuples, reverse=True)


def mac_get_crossover_bottles():
    bottles_path = os.path.expanduser('~/Library/Application Support/CrossOver/Bottles')
    if not os.path.exists(bottles_path):
        return []
    return sorted(p for p in os.listdir(bottles_path) if mac_is_valid_bottle(p))


def mac_is_valid_bottle(bottle_name):
    bottles_path = os.path.expanduser('~/Library/Application Support/CrossOver/Bottles')
    return os.path.exists(os.path.join(bottles_path, bottle_name, 'cxbottle.conf'))


def mac_get_bottle_path(bottle_name):
    bottles_path = os.path.expanduser('~/Library/Application Support/CrossOver/Bottles')
    return os.path.join(bottles_path, bottle_name)


def mac_is_crossover_running():
    try:
        out = subprocess.check_output(['launchctl', 'list'])
        return b'com.codeweavers.CrossOver.' in out
    except Exception as e:
        _logger.warning(f'Getting list of running application bundles failed: {e!r}')
        return True  # assume the worst



# coding: utf-8

from datetime import datetime
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, List, Dict


@dataclass
class GameAsset:
    """
    App asset data
    """
    app_name: str = ''
    asset_id: str = ''
    build_version: str = ''
    catalog_item_id: str = ''
    label_name: str = ''
    namespace: str = ''
    metadata: Dict = field(default_factory=dict)

    @classmethod
    def from_egs_json(cls, json):
        tmp = cls()
        tmp.app_name = json.get('appName', '')
        tmp.asset_id = json.get('assetId', '')
        tmp.build_version = json.get('buildVersion', '')
        tmp.catalog_item_id = json.get('catalogItemId', '')
        tmp.label_name = json.get('labelName', '')
        tmp.namespace = json.get('namespace', '')
        tmp.metadata = json.get('metadata', {})
        return tmp

    @classmethod
    def from_json(cls, json):
        tmp = cls()
        tmp.app_name = json.get('app_name', '')
        tmp.asset_id = json.get('asset_id', '')
        tmp.build_version = json.get('build_version', '')
        tmp.catalog_item_id = json.get('catalog_item_id', '')
        tmp.label_name = json.get('label_name', '')
        tmp.namespace = json.get('namespace', '')
        tmp.metadata = json.get('metadata', {})
        return tmp


@dataclass
class Game:
    """
    Combination of app asset and app metadata as stored on disk
    """
    app_name: str
    app_title: str

    asset_infos: Dict[str, GameAsset] = field(default_factory=dict)
    base_urls: List[str] = field(default_factory=list)
    metadata: Dict = field(default_factory=dict)

    def app_version(self, platform='Windows'):
        if platform not in self.asset_infos:
            return None
        return self.asset_infos[platform].build_version

    @property
    def is_dlc(self):
        return self.metadata and 'mainGameItem' in self.metadata

    @property
    def third_party_store(self):
        if not self.metadata:
            return None
        return self.metadata.get('customAttributes', {}).get('ThirdPartyManagedApp', {}).get('value', None)

    @property
    def partner_link_type(self):
        if not self.metadata:
            return None
        return self.metadata.get('customAttributes', {}).get('partnerLinkType', {}).get('value', None)

    @property
    def partner_link_id(self):
        if not self.metadata:
            return None
        return self.metadata.get('customAttributes', {}).get('partnerLinkId', {}).get('value', None)

    @property
    def supports_cloud_saves(self):
        return self.metadata and (self.metadata.get('customAttributes', {}).get('CloudSaveFolder') is not None)

    @property
    def supports_mac_cloud_saves(self):
        return self.metadata and (self.metadata.get('customAttributes', {}).get('CloudSaveFolder_MAC') is not None)

    @property
    def catalog_item_id(self):
        if not self.metadata:
            return None
        return self.metadata['id']

    @property
    def namespace(self):
        if not self.metadata:
            return None
        return self.metadata['namespace']

    @classmethod
    def from_json(cls, json):
        tmp = cls(
            app_name=json.get('app_name', ''),
            app_title=json.get('app_title', ''),
        )
        tmp.metadata = json.get('metadata', dict())
        if 'asset_infos' in json:
            tmp.asset_infos = {k: GameAsset.from_json(v) for k, v in json['asset_infos'].items()}
        else:
            # Migrate old asset_info to new asset_infos
            tmp.asset_infos['Windows'] = GameAsset.from_json(json.get('asset_info', dict()))

        tmp.base_urls = json.get('base_urls', list())
        return tmp

    @property
    def __dict__(self):
        """This is just here so asset_infos gets turned into a dict as well"""
        assets_dictified = {k: v.__dict__ for k, v in self.asset_infos.items()}
        return dict(metadata=self.metadata, asset_infos=assets_dictified, app_name=self.app_name,
                    app_title=self.app_title, base_urls=self.base_urls)


@dataclass
class InstalledGame:
    """
    Local metadata for an installed app
    """
    app_name: str
    install_path: str
    title: str
    version: str

    base_urls: List[str] = field(default_factory=list)
    can_run_offline: bool = False
    egl_guid: str = ''
    executable: str = ''
    install_size: int = 0
    install_tags: List[str] = field(default_factory=list)
    is_dlc: bool = False
    launch_parameters: str = ''
    manifest_path: str = ''
    needs_verification: bool = False
    platform: str = 'Windows'
    prereq_info: Optional[Dict] = None
    uninstaller: Optional[Dict] = None
    requires_ot: bool = False
    save_path: Optional[str] = None

    @classmethod
    def from_json(cls, json):
        tmp = cls(
            app_name=json.get('app_name', ''),
            install_path=json.get('install_path', ''),
            title=json.get('title', ''),
            version=json.get('version', ''),
        )

        tmp.base_urls = json.get('base_urls', list())
        tmp.executable = json.get('executable', '')
        tmp.launch_parameters = json.get('launch_parameters', '')
        tmp.prereq_info = json.get('prereq_info', None)
        tmp.uninstaller = json.get('uninstaller', None)

        tmp.can_run_offline = json.get('can_run_offline', False)
        tmp.requires_ot = json.get('requires_ot', False)
        tmp.is_dlc = json.get('is_dlc', False)
        tmp.save_path = json.get('save_path', None)
        tmp.manifest_path = json.get('manifest_path', '')
        tmp.needs_verification = json.get('needs_verification', False) is True
        tmp.platform = json.get('platform', 'Windows')
        tmp.install_size = json.get('install_size', 0)
        tmp.egl_guid = json.get('egl_guid', '')
        tmp.install_tags = json.get('install_tags', [])
        return tmp


@dataclass
class SaveGameFile:
    """
    Metadata for a cloud save manifest
    """
    app_name: str
    filename: str
    manifest_name: str
    datetime: Optional[datetime] = None


class SaveGameStatus(Enum):
    LOCAL_NEWER = 0
    REMOTE_NEWER = 1
    SAME_AGE = 2
    NO_SAVE = 3


class VerifyResult(Enum):
    HASH_MATCH = 0
    HASH_MISMATCH = 1
    FILE_MISSING = 2
    OTHER_ERROR = 3


@dataclass
class LaunchParameters:
    """
    Parameters for launching a game
    """
    # game-supplied parameters
    game_parameters: list = field(default_factory=list)
    game_executable: str = ''
    game_directory: str = ''
    # EGL parameters (auth, ovt, etc.)
    egl_parameters: list = field(default_factory=list)
    # command line before executable (WINE, gamemode, etc.)
    launch_command: list = field(default_factory=list)
    # working directory for launched process
    working_directory: str = ''
    # user and environment supplied options
    user_parameters: list = field(default_factory=list)
    environment: dict = field(default_factory=dict)
    pre_launch_command: str = ''
    pre_launch_wait: bool = False



# coding: utf-8

import json
import os
import logging

from contextlib import contextmanager
from collections import defaultdict
from pathlib import Path
from time import time

from filelock import FileLock

from .utils import clean_filename, LockedJSONData

from legendary.models.game import *
from legendary.utils.aliasing import generate_aliases
from legendary.models.config import LGDConf
from legendary.utils.env import is_windows_mac_or_pyi


FILELOCK_DEBUG = False


class LGDLFS:
    def __init__(self, config_file=None):
        self.log = logging.getLogger('LGDLFS')

        if config_path := os.environ.get('LEGENDARY_CONFIG_PATH'):
            self.path = config_path
        elif config_path := os.environ.get('XDG_CONFIG_HOME'):
            self.path = os.path.join(config_path, 'legendary')
        else:
            self.path = os.path.expanduser('~/.config/legendary')

        # EGS user info
        self._user_data = None
        # EGS entitlements
        self._entitlements = None
        # EGS asset data
        self._assets = None
        # EGS metadata
        self._game_metadata = dict()
        # Legendary update check info
        self._update_info = None
        # EOS Overlay install/update check info
        self._overlay_update_info = None
        self._overlay_install_info = None
        # Config with game specific settings (e.g. start parameters, env variables)
        self.config = LGDConf(comment_prefixes='/', allow_no_value=True)

        if config_file:
            # if user specified a valid relative/absolute path use that,
            # otherwise create file in legendary config directory
            if os.path.exists(config_file):
                self.config_path = os.path.abspath(config_file)
            else:
                self.config_path = os.path.join(self.path, clean_filename(config_file))
            self.log.info(f'Using non-default config file "{self.config_path}"')
        else:
            self.config_path = os.path.join(self.path, 'config.ini')

        # ensure folders exist.
        for f in ['', 'manifests', 'metadata', 'tmp']:
            if not os.path.exists(os.path.join(self.path, f)):
                os.makedirs(os.path.join(self.path, f))

        # if "old" folder exists migrate files and remove it
        if os.path.exists(os.path.join(self.path, 'manifests', 'old')):
            self.log.info('Migrating manifest files from old folders to new, please wait...')
            # remove unversioned manifest files
            for _f in os.listdir(os.path.join(self.path, 'manifests')):
                if '.manifest' not in _f:
                    continue
                if '_' not in _f or (_f.startswith('UE_') and _f.count('_') < 2):
                    self.log.debug(f'Deleting "{_f}" ...')
                    os.remove(os.path.join(self.path, 'manifests', _f))

            # move files from "old" to the base folder
            for _f in os.listdir(os.path.join(self.path, 'manifests', 'old')):
                try:
                    self.log.debug(f'Renaming "{_f}"')
                    os.rename(os.path.join(self.path, 'manifests', 'old', _f),
                              os.path.join(self.path, 'manifests', _f))
                except Exception as e:
                    self.log.warning(f'Renaming manifest file "{_f}" failed: {e!r}')

            # remove "old" folder
            try:
                os.removedirs(os.path.join(self.path, 'manifests', 'old'))
            except Exception as e:
                self.log.warning(f'Removing "{os.path.join(self.path, "manifests", "old")}" folder failed: '
                                 f'{e!r}, please remove manually')

        if not FILELOCK_DEBUG:
            # Prevent filelock logger from spamming Legendary debug output
            filelock_logger = logging.getLogger('filelock')
            filelock_logger.setLevel(logging.INFO)

        # try loading config
        try:
            self.config.read(self.config_path)
        except Exception as e:
            self.log.error(f'Unable to read configuration file, please ensure that file is valid! '
                           f'(Error: {repr(e)})')
            self.log.warning('Continuing with blank config in safe-mode...')
            self.config.read_only = True

        # make sure "Legendary" section exists
        if 'Legendary' not in self.config:
            self.config.add_section('Legendary')

        # Add opt-out options with explainers
        if not self.config.has_option('Legendary', 'disable_update_check'):
            self.config.set('Legendary', '; Disables the automatic update check')
            self.config.set('Legendary', 'disable_update_check', 'false')
        if not self.config.has_option('Legendary', 'disable_update_notice'):
            self.config.set('Legendary', '; Disables the notice about an available update on exit')
            self.config.set('Legendary', 'disable_update_notice', 'false' if is_windows_mac_or_pyi() else 'true')

        self._installed_lock = FileLock(os.path.join(self.path, 'installed.json') + '.lock')

        try:
            self._installed = json.load(open(os.path.join(self.path, 'installed.json')))
        except Exception as e:
            self.log.debug(f'Loading installed games failed: {e!r}')
            self._installed = None

        # load existing app metadata
        for gm_file in os.listdir(os.path.join(self.path, 'metadata')):
            try:
                _meta = json.load(open(os.path.join(self.path, 'metadata', gm_file)))
                self._game_metadata[_meta['app_name']] = _meta
            except Exception as e:
                self.log.debug(f'Loading game meta file "{gm_file}" failed: {e!r}')

        # load auto-aliases if enabled
        self.aliases = dict()
        if not self.config.getboolean('Legendary', 'disable_auto_aliasing', fallback=False):
            try:
                _j = json.load(open(os.path.join(self.path, 'aliases.json')))
                for app_name, aliases in _j.items():
                    for alias in aliases:
                        self.aliases[alias] = app_name
            except Exception as e:
                self.log.debug(f'Loading aliases failed with {e!r}')

    @property
    @contextmanager
    def userdata_lock(self) -> LockedJSONData:
        """Wrapper around the lock to automatically update user data when it is released"""
        with LockedJSONData(os.path.join(self.path, 'user.json')) as lock:
            try:
                yield lock
            finally:
                self._user_data = lock.data

    @property
    def userdata(self):
        if self._user_data is not None:
            return self._user_data

        try:
            with self.userdata_lock as locked:
                return locked.data
        except Exception as e:
            self.log.debug(f'Failed to load user data: {e!r}')
            return None

    @userdata.setter
    def userdata(self, userdata):
        raise NotImplementedError('The setter has been removed, use the locked userdata instead.')

    def invalidate_userdata(self):
        with self.userdata_lock as lock:
            lock.clear()

    @property
    def entitlements(self):
        if self._entitlements is not None:
            return self._entitlements

        try:
            self._entitlements = json.load(open(os.path.join(self.path, 'entitlements.json')))
            return self._entitlements
        except Exception as e:
            self.log.debug(f'Failed to load entitlements data: {e!r}')
            return None

    @entitlements.setter
    def entitlements(self, entitlements):
        if entitlements is None:
            raise ValueError('Entitlements is none!')

        self._entitlements = entitlements
        json.dump(entitlements, open(os.path.join(self.path, 'entitlements.json'), 'w'),
                  indent=2, sort_keys=True)

    @property
    def assets(self):
        if self._assets is None:
            try:
                tmp = json.load(open(os.path.join(self.path, 'assets.json')))
                self._assets = {k: [GameAsset.from_json(j) for j in v] for k, v in tmp.items()}
            except Exception as e:
                self.log.debug(f'Failed to load assets data: {e!r}')
                return None

        return self._assets

    @assets.setter
    def assets(self, assets):
        if assets is None:
            raise ValueError('Assets is none!')

        self._assets = assets
        json.dump({platform: [a.__dict__ for a in assets] for platform, assets in self._assets.items()},
                  open(os.path.join(self.path, 'assets.json'), 'w'),
                  indent=2, sort_keys=True)

    def _get_manifest_filename(self, app_name, version, platform=None):
        if platform:
            fname = clean_filename(f'{app_name}_{platform}_{version}')
        else:
            fname = clean_filename(f'{app_name}_{version}')
        return os.path.join(self.path, 'manifests', f'{fname}.manifest')

    def load_manifest(self, app_name, version, platform='Windows'):
        try:
            return open(self._get_manifest_filename(app_name, version, platform), 'rb').read()
        except FileNotFoundError:  # all other errors should propagate
            self.log.debug(f'Loading manifest failed, retrying without platform in filename...')
            try:
                return open(self._get_manifest_filename(app_name, version), 'rb').read()
            except FileNotFoundError:  # all other errors should propagate
                return None

    def save_manifest(self, app_name, manifest_data, version, platform='Windows'):
        with open(self._get_manifest_filename(app_name, version, platform), 'wb') as f:
            f.write(manifest_data)

    def get_game_meta(self, app_name):
        if _meta := self._game_metadata.get(app_name, None):
            return Game.from_json(_meta)
        return None

    def set_game_meta(self, app_name, meta):
        json_meta = meta.__dict__
        self._game_metadata[app_name] = json_meta
        meta_file = os.path.join(self.path, 'metadata', f'{app_name}.json')
        json.dump(json_meta, open(meta_file, 'w'), indent=2, sort_keys=True)

    def delete_game_meta(self, app_name):
        if app_name not in self._game_metadata:
            raise ValueError(f'Game {app_name} does not exist in metadata DB!')

        del self._game_metadata[app_name]
        meta_file = os.path.join(self.path, 'metadata', f'{app_name}.json')
        if os.path.exists(meta_file):
            os.remove(meta_file)

    def get_game_app_names(self):
        return sorted(self._game_metadata.keys())

    def get_tmp_path(self):
        return os.path.join(self.path, 'tmp')

    def clean_tmp_data(self):
        for f in os.listdir(os.path.join(self.path, 'tmp')):
            try:
                os.remove(os.path.join(self.path, 'tmp', f))
            except Exception as e:
                self.log.warning(f'Failed to delete file "{f}": {e!r}')

    def clean_metadata(self, app_names):
        for f in os.listdir(os.path.join(self.path, 'metadata')):
            app_name = f.rpartition('.')[0]
            if app_name not in app_names:
                try:
                    os.remove(os.path.join(self.path, 'metadata', f))
                except Exception as e:
                    self.log.warning(f'Failed to delete file "{f}": {e!r}')

    def clean_manifests(self, in_use):
        in_use_files = {
            f'{clean_filename(f"{app_name}_{version}")}.manifest'
            for app_name, version, _ in in_use
        }

        in_use_files |= {
            f'{clean_filename(f"{app_name}_{platform}_{version}")}.manifest'
            for app_name, version, platform in in_use
        }

        for f in os.listdir(os.path.join(self.path, 'manifests')):
            if f not in in_use_files:
                try:
                    os.remove(os.path.join(self.path, 'manifests', f))
                except Exception as e:
                    self.log.warning(f'Failed to delete file "{f}": {e!r}')

    def lock_installed(self) -> bool:
        """
        Locks the install data. We do not care about releasing this lock.
        If it is acquired by a Legendary instance it should own the lock until it exits.
        Some operations such as egl sync may be simply skipped if a lock cannot be acquired
        """
        if self._installed_lock.is_locked:
            return True

        try:
            self._installed_lock.acquire(blocking=False)
            # reload data in case it has been updated elsewhere
            try:
                self._installed = json.load(open(os.path.join(self.path, 'installed.json')))
            except Exception as e:
                self.log.debug(f'Failed to load installed game data: {e!r}')

            return True
        except TimeoutError:
            return False

    def get_installed_game(self, app_name):
        if self._installed is None:
            try:
                self._installed = json.load(open(os.path.join(self.path, 'installed.json')))
            except Exception as e:
                self.log.debug(f'Failed to load installed game data: {e!r}')
                return None

        if game_json := self._installed.get(app_name, None):
            return InstalledGame.from_json(game_json)
        return None

    def set_installed_game(self, app_name, install_info):
        if self._installed is None:
            self._installed = dict()

        if app_name in self._installed:
            self._installed[app_name].update(install_info.__dict__)
        else:
            self._installed[app_name] = install_info.__dict__

        json.dump(self._installed, open(os.path.join(self.path, 'installed.json'), 'w'),
                  indent=2, sort_keys=True)

    def remove_installed_game(self, app_name):
        if self._installed is None:
            self.log.warning('Trying to remove a game, but no installed games?!')
            return

        if app_name in self._installed:
            del self._installed[app_name]
        else:
            self.log.warning('Trying to remove non-installed game:', app_name)
            return

        json.dump(self._installed, open(os.path.join(self.path, 'installed.json'), 'w'),
                  indent=2, sort_keys=True)

    def get_installed_list(self):
        if not self._installed:
            return []

        return [InstalledGame.from_json(i) for i in self._installed.values()]

    def save_config(self):
        # do not save if in read-only mode or file hasn't changed
        if self.config.read_only or not self.config.modified:
            return
        # if config file has been modified externally, back-up the user-modified version before writing
        if os.path.exists(self.config_path):
            if (modtime := int(os.stat(self.config_path).st_mtime)) != self.config.modtime:
                new_filename = f'config.{modtime}.ini'
                self.log.warning(f'Configuration file has been modified while legendary was running, '
                                 f'user-modified config will be renamed to "{new_filename}"...')
                os.rename(self.config_path, os.path.join(os.path.dirname(self.config_path), new_filename))

        with open(self.config_path, 'w') as cf:
            self.config.write(cf)

    def get_dir_size(self):
        return sum(f.stat().st_size for f in Path(self.path).glob('**/*') if f.is_file())

    def get_cached_version(self):
        if self._update_info:
            return self._update_info

        try:
            self._update_info = json.load(open(os.path.join(self.path, 'version.json')))
        except Exception as e:
            self.log.debug(f'Failed to load cached update data: {e!r}')
            self._update_info = dict(last_update=0, data=None)

        return self._update_info

    def set_cached_version(self, version_data):
        if not version_data:
            return
        self._update_info = dict(last_update=time(), data=version_data)
        json.dump(self._update_info, open(os.path.join(self.path, 'version.json'), 'w'),
                  indent=2, sort_keys=True)

    def get_cached_sdl_data(self, app_name):
        try:
            return json.load(open(os.path.join(self.path, 'tmp', f'{app_name}.json')))
        except Exception as e:
            self.log.debug(f'Failed to load cached SDL data: {e!r}')
            return None

    def set_cached_sdl_data(self, app_name, sdl_version, sdl_data):
        if not app_name or not sdl_data:
            return
        json.dump(dict(version=sdl_version, data=sdl_data),
                  open(os.path.join(self.path, 'tmp', f'{app_name}.json'), 'w'),
                  indent=2, sort_keys=True)

    def get_cached_overlay_version(self):
        if self._overlay_update_info:
            return self._overlay_update_info

        try:
            self._overlay_update_info = json.load(open(
                os.path.join(self.path, 'overlay_version.json')))
        except Exception as e:
            self.log.debug(f'Failed to load cached Overlay update data: {e!r}')
            self._overlay_update_info = dict(last_update=0, data=None)

        return self._overlay_update_info

    def set_cached_overlay_version(self, version_data):
        self._overlay_update_info = dict(last_update=time(), data=version_data)
        json.dump(self._overlay_update_info,
                  open(os.path.join(self.path, 'overlay_version.json'), 'w'),
                  indent=2, sort_keys=True)

    def get_overlay_install_info(self):
        if not self._overlay_install_info:
            try:
                data = json.load(open(os.path.join(self.path, 'overlay_install.json')))
                self._overlay_install_info = InstalledGame.from_json(data)
            except Exception as e:
                self.log.debug(f'Failed to load overlay install data: {e!r}')

        return self._overlay_install_info

    def set_overlay_install_info(self, igame: InstalledGame):
        self._overlay_install_info = igame
        json.dump(vars(igame), open(os.path.join(self.path, 'overlay_install.json'), 'w'),
                  indent=2, sort_keys=True)

    def remove_overlay_install_info(self):
        try:
            self._overlay_install_info = None
            os.remove(os.path.join(self.path, 'overlay_install.json'))
        except Exception as e:
            self.log.debug(f'Failed to delete overlay install data: {e!r}')

    def generate_aliases(self):
        self.log.debug('Generating list of aliases...')

        self.aliases = dict()
        aliases = set()
        collisions = set()
        alias_map = defaultdict(set)

        for app_name in self._game_metadata.keys():
            game = self.get_game_meta(app_name)
            if game.is_dlc:
                continue
            game_folder = game.metadata.get('customAttributes', {}).get('FolderName', {}).get('value', None)
            _aliases = generate_aliases(game.app_title, game_folder=game_folder, app_name=game.app_name)
            for alias in _aliases:
                if alias not in aliases:
                    aliases.add(alias)
                    alias_map[game.app_name].add(alias)
                else:
                    collisions.add(alias)

        # remove colliding aliases from map and add aliases to lookup table
        for app_name, aliases in alias_map.items():
            alias_map[app_name] -= collisions
            for alias in alias_map[app_name]:
                self.aliases[alias] = app_name

        def serialise_sets(obj):
            """Turn sets into sorted lists for storage"""
            return sorted(obj) if isinstance(obj, set) else obj

        json.dump(alias_map, open(os.path.join(self.path, 'aliases.json'), 'w', newline='\n'),
                  indent=2, sort_keys=True, default=serialise_sets)



from copy import deepcopy

from legendary.models.game import InstalledGame, Game
from legendary.utils.cli import strtobool


_template = {
    'AppCategories': ['public', 'games', 'applications'],
    'AppName': '',
    'AppVersionString': '',
    'BaseURLs': [],
    'BuildLabel': '',
    'CatalogItemId': '',
    'CatalogNamespace': '',
    'ChunkDbs': [],
    'CompatibleApps': [],
    'DisplayName': '',
    'FormatVersion': 0,
    'FullAppName': '',
    'HostInstallationGuid': '',
    'InstallComponents': [],
    'InstallLocation': '',
    'InstallSessionId': '',
    'InstallSize': 0,
    'InstallTags': [],
    'InstallationGuid': '',
    'LaunchCommand': '',
    'LaunchExecutable': '',
    'MainGameAppName': '',
    'MainWindowProcessName': '',
    'MandatoryAppFolderName': '',
    'ManifestLocation': '',
    'OwnershipToken': '',
    'PrereqIds': [],
    'ProcessNames': [],
    'StagingLocation': '',
    'TechnicalType': '',
    'VaultThumbnailUrl': '',
    'VaultTitleText': '',
    'bCanRunOffline': True,
    'bIsApplication': True,
    'bIsExecutable': True,
    'bIsIncompleteInstall': False,
    'bIsManaged': False,
    'bNeedsValidation': False,
    'bRequiresAuth': True
}


class EGLManifest:
    def __init__(self):
        self.app_name = None
        self.app_version_string = None
        self.base_urls = None
        self.build_label = None
        self.catalog_item_id = None
        self.namespace = None
        self.display_name = None
        self.install_location = None
        self.install_size = None
        self.install_tags = None
        self.installation_guid = None
        self.launch_command = None
        self.executable = None
        self.main_game_appname = None
        self.app_folder_name = None
        self.manifest_location = None
        self.ownership_token = None
        self.staging_location = None
        self.can_run_offline = None
        self.is_incomplete_install = None
        self.needs_validation = None

        self.remainder = dict()

    @classmethod
    def from_json(cls, json: dict):
        json = deepcopy(json)
        tmp = cls()
        tmp.app_name = json.pop('AppName')
        tmp.app_version_string = json.pop('AppVersionString', None)
        tmp.base_urls = json.pop('BaseURLs', list())
        tmp.build_label = json.pop('BuildLabel', '')
        tmp.catalog_item_id = json.pop('CatalogItemId', '')
        tmp.namespace = json.pop('CatalogNamespace', '')
        tmp.display_name = json.pop('DisplayName', '')
        tmp.install_location = json.pop('InstallLocation', '')
        tmp.install_size = json.pop('InstallSize', 0)
        tmp.install_tags = json.pop('InstallTags', [])
        tmp.installation_guid = json.pop('InstallationGuid', '')
        tmp.launch_command = json.pop('LaunchCommand', '')
        tmp.executable = json.pop('LaunchExecutable', '')
        tmp.main_game_appname = json.pop('MainGameAppName', '')
        tmp.app_folder_name = json.pop('MandatoryAppFolderName', '')
        tmp.manifest_location = json.pop('ManifestLocation', '')
        tmp.ownership_token = strtobool(json.pop('OwnershipToken', 'False'))
        tmp.staging_location = json.pop('StagingLocation', '')
        tmp.can_run_offline = json.pop('bCanRunOffline', True)
        tmp.is_incomplete_install = json.pop('bIsIncompleteInstall', False)
        tmp.needs_validation = json.pop('bNeedsValidation', False)
        tmp.remainder = json.copy()
        return tmp

    def to_json(self) -> dict:
        out = _template.copy()
        out.update(self.remainder)
        out['AppName'] = self.app_name
        out['AppVersionString'] = self.app_version_string
        out['BaseURLs'] = self.base_urls
        out['BuildLabel'] = self.build_label
        out['CatalogItemId'] = self.catalog_item_id
        out['CatalogNamespace'] = self.namespace
        out['DisplayName'] = self.display_name
        out['InstallLocation'] = self.install_location
        out['InstallSize'] = self.install_size
        out['InstallTags'] = self.install_tags
        out['InstallationGuid'] = self.installation_guid
        out['LaunchCommand'] = self.launch_command
        out['LaunchExecutable'] = self.executable
        out['MainGameAppName'] = self.main_game_appname
        out['MandatoryAppFolderName'] = self.app_folder_name
        out['ManifestLocation'] = self.manifest_location
        out['OwnershipToken'] = str(self.ownership_token).lower()
        out['StagingLocation'] = self.staging_location
        out['bCanRunOffline'] = self.can_run_offline
        out['bIsIncompleteInstall'] = self.is_incomplete_install
        out['bNeedsValidation'] = self.needs_validation
        return out

    @classmethod
    def from_lgd_game(cls, game: Game, igame: InstalledGame):
        tmp = cls()
        tmp.app_name = game.app_name
        tmp.app_version_string = igame.version
        tmp.base_urls = igame.base_urls
        tmp.build_label = 'Live'
        tmp.catalog_item_id = game.catalog_item_id
        tmp.namespace = game.namespace
        tmp.display_name = igame.title
        tmp.install_location = igame.install_path
        tmp.install_size = igame.install_size
        tmp.install_tags = igame.install_tags
        tmp.installation_guid = igame.egl_guid
        tmp.launch_command = igame.launch_parameters
        tmp.executable = igame.executable
        tmp.main_game_appname = game.app_name  # todo for DLC support this needs to be the base game
        tmp.app_folder_name = game.metadata.get('customAttributes', {}).get('FolderName', {}).get('value', '')
        tmp.manifest_location = f'{igame.install_path}/.egstore'
        tmp.ownership_token = igame.requires_ot
        tmp.staging_location = f'{igame.install_path}/.egstore/bps'
        tmp.can_run_offline = igame.can_run_offline
        tmp.is_incomplete_install = False
        tmp.needs_validation = igame.needs_verification
        return tmp

    def to_lgd_igame(self) -> InstalledGame:
        return InstalledGame(app_name=self.app_name, title=self.display_name, version=self.app_version_string,
                             base_urls=self.base_urls, install_path=self.install_location, executable=self.executable,
                             launch_parameters=self.launch_command, can_run_offline=self.can_run_offline,
                             requires_ot=self.ownership_token, is_dlc=False,
                             needs_verification=self.needs_validation, install_size=self.install_size,
                             egl_guid=self.installation_guid, install_tags=self.install_tags)



import os
import logging

from legendary.models.game import Game

if os.name == 'nt':
    from legendary.lfs.windows_helpers import *

logger = logging.getLogger('EOSUtils')
# Dummy Game objects to use with Core methods that expect them
# Overlay
EOSOverlayApp = Game(app_name='98bc04bc842e4906993fd6d6644ffb8d',
                     app_title='Epic Online Services Overlay',
                     metadata=dict(namespace='302e5ede476149b1bc3e4fe6ae45e50e',
                                   id='cc15684f44d849e89e9bf4cec0508b68'))
# EOS Windows service
EOSHApp = Game(app_name='c9e2eb9993a1496c99dc529b49a07339',
               app_title='Epic Online Services Helper (EOSH)',
               metadata=dict(namespace='302e5ede476149b1bc3e4fe6ae45e50e',
                             id='1108a9c0af47438da91331753b22ea21'))

EOS_OVERLAY_KEY = r'SOFTWARE\Epic Games\EOS'
WINE_EOS_OVERLAY_KEY = EOS_OVERLAY_KEY.replace('\\', '\\\\')
EOS_OVERLAY_VALUE = 'OverlayPath'
VULKAN_OVERLAY_KEY = r'SOFTWARE\Khronos\Vulkan\ImplicitLayers'


def query_registry_entries(prefix=None):
    if os.name == 'nt':
        # Overlay location for the EOS SDK to load
        overlay_path = query_registry_value(HKEY_CURRENT_USER, EOS_OVERLAY_KEY, EOS_OVERLAY_VALUE)
        # Vulkan Layers
        # HKCU
        vulkan_hkcu = [i[0] for i in
                       list_registry_values(HKEY_CURRENT_USER, VULKAN_OVERLAY_KEY)
                       if 'EOS' in i[0]]
        # HKLM 64 & 32 bit
        vulkan_hklm = [i[0] for i in
                       list_registry_values(HKEY_LOCAL_MACHINE, VULKAN_OVERLAY_KEY)
                       if 'EOS' in i[0]]
        vulkan_hklm += [i[0] for i in
                        list_registry_values(HKEY_LOCAL_MACHINE, VULKAN_OVERLAY_KEY, use_32bit_view=True)
                        if 'EOS' in i[0]]

        return dict(overlay_path=overlay_path,
                    vulkan_hkcu=vulkan_hkcu,
                    vulkan_hklm=vulkan_hklm)
    elif prefix:
        # Only read HKCU since we don't really care for the Vulkan stuff (doesn't work in WINE)
        use_reg_file = os.path.join(prefix, 'user.reg')
        if not os.path.exists(use_reg_file):
            raise ValueError('No user.reg file, invalid path')

        reg_lines = open(use_reg_file, 'r', encoding='utf-8').readlines()
        for line in reg_lines:
            if EOS_OVERLAY_VALUE in line:
                overlay_path = line.partition('=')[2].strip().strip('"')
                break
        else:
            overlay_path = None

        if overlay_path:
            if overlay_path.startswith('C:'):
                overlay_path = os.path.join(prefix, 'drive_c', overlay_path[3:])
            elif overlay_path.startswith('Z:'):
                overlay_path = overlay_path[2:]

        return dict(overlay_path=overlay_path,
                    vulkan_hkcu=list(),
                    vulkan_hklm=list())
    else:
        raise ValueError('No prefix specified on non-Windows platform')


def add_registry_entries(overlay_path, prefix=None):
    if os.name == 'nt':
        logger.debug(f'Settings HKCU EOS Overlay Path: {overlay_path}')
        set_registry_value(HKEY_CURRENT_USER, EOS_OVERLAY_KEY, EOS_OVERLAY_VALUE,
                           overlay_path.replace('\\', '/'), TYPE_STRING)
        vk_32_path = os.path.join(overlay_path, 'EOSOverlayVkLayer-Win32.json').replace('/', '\\')
        vk_64_path = os.path.join(overlay_path, 'EOSOverlayVkLayer-Win64.json').replace('/', '\\')
        # the launcher only sets those in HKCU, th e service sets them in HKLM,
        # but it's not in use yet, so just do HKCU for now
        logger.debug(f'Settings HKCU 32-bit Vulkan Layer: {vk_32_path}')
        set_registry_value(HKEY_CURRENT_USER, VULKAN_OVERLAY_KEY, vk_32_path, 0, TYPE_DWORD)
        logger.debug(f'Settings HKCU 64-bit Vulkan Layer: {vk_32_path}')
        set_registry_value(HKEY_CURRENT_USER, VULKAN_OVERLAY_KEY, vk_64_path, 0, TYPE_DWORD)
    elif prefix:
        # Again only care for HKCU OverlayPath because Windows Vulkan layers don't work anyway
        use_reg_file = os.path.join(prefix, 'user.reg')
        if not os.path.exists(use_reg_file):
            raise ValueError('No user.reg file, invalid path')

        reg_lines = open(use_reg_file, 'r', encoding='utf-8').readlines()

        overlay_path = overlay_path.replace('\\', '/')
        if overlay_path.startswith('/'):
            overlay_path = f'Z:{overlay_path}'

        overlay_line = f'"{EOS_OVERLAY_VALUE}"="{overlay_path}"\n'
        overlay_idx = None
        section_idx = None

        for idx, line in enumerate(reg_lines):
            if EOS_OVERLAY_VALUE in line:
                reg_lines[idx] = overlay_line
                break
            elif WINE_EOS_OVERLAY_KEY in line:
                section_idx = idx
        else:
            if section_idx:
                reg_lines.insert(section_idx + 1, overlay_line)
            else:
                reg_lines.append(f'[{WINE_EOS_OVERLAY_KEY}]\n')
                reg_lines.append(overlay_line)

        open(use_reg_file, 'w', encoding='utf-8').writelines(reg_lines)
    else:
        raise ValueError('No prefix specified on non-Windows platform')


def remove_registry_entries(prefix=None):
    entries = query_registry_entries(prefix)

    if os.name == 'nt':
        if entries['overlay_path']:
            logger.debug('Removing HKCU EOS OverlayPath')
            remove_registry_value(HKEY_CURRENT_USER, EOS_OVERLAY_KEY, EOS_OVERLAY_VALUE)
        for value in entries['vulkan_hkcu']:
            logger.debug(f'Removing HKCU Vulkan Layer: {value}')
            remove_registry_value(HKEY_CURRENT_USER, VULKAN_OVERLAY_KEY, value)
        for value in entries['vulkan_hklm']:
            logger.debug(f'Removing HKLM Vulkan Layer: {value}')
            remove_registry_value(HKEY_LOCAL_MACHINE, VULKAN_OVERLAY_KEY, value)
            remove_registry_value(HKEY_LOCAL_MACHINE, VULKAN_OVERLAY_KEY, value, use_32bit_view=True)
    elif prefix:
        # Same as above, only HKCU.
        use_reg_file = os.path.join(prefix, 'user.reg')
        if not os.path.exists(use_reg_file):
            raise ValueError('No user.reg file, invalid path')

        if entries['overlay_path']:
            reg_lines = open(use_reg_file, 'r', encoding='utf-8').readlines()
            filtered_lines = [line for line in reg_lines if EOS_OVERLAY_VALUE not in line]
            open(use_reg_file, 'w', encoding='utf-8').writelines(filtered_lines)
    else:
        raise ValueError('No prefix specified on non-Windows platform')



from string import ascii_lowercase, digits

# Aliases generated:
# - name lowercase (without TM etc.)
# - same, but without spaces
# - same, but roman numerals are replaced
# if name has >= 2 parts:
# - initials
# - initials, but roman numerals are intact
# - initials, but roman numerals are replaced with number
# if ':' in name:
# - run previous recursively with everything before ":"
# if single 'f' in long word:
# - split word (this is mainly for cases like Battlfront -> BF)
# the first word longer than 1 character that isn't "the", "for", or "of" will also be added

allowed_characters = ascii_lowercase+digits
roman = {
    'i': '1',
    'ii': '2',
    'iii': '3',
    'iv': '4',
    'v': '5',
    'vi': '6',
    'vii': '7',
    'viii': '8',
    'ix': '9',
    'x': '10',
    'xi': '11',
    'xii': '12',
    'xiii': '13',
    'xiv': '14',
    'xv': '15',
    'xvi': '16',
    'xvii': '17',
    'xviii': '18',
    'xix': '19',
    'xx': '20'
}


def _filter(input):
    return ''.join(l for l in input if l in allowed_characters)


def generate_aliases(game_name, game_folder=None, split_words=True, app_name=None):
    # normalise and split name, then filter for legal characters
    game_parts = [_filter(p) for p in game_name.lower().split()]
    # filter out empty parts
    game_parts = [p for p in game_parts if p]

    _aliases = [
        game_name.lower().strip(),
        ' '.join(game_parts),
        ''.join(game_parts),
        ''.join(roman.get(p, p) for p in game_parts),
    ]

    # single word abbreviation
    try:
        first_word = next(i for i in game_parts if i not in ('for', 'the', 'of'))
        if len(first_word) > 1:
            _aliases.append(first_word)
    except StopIteration:
        pass

    # remove subtitle from game
    if ':' in game_name:
        _aliases.extend(generate_aliases(game_name.partition(':')[0]))
    if '-' in game_name:
        _aliases.extend(generate_aliases(game_name.replace('-', ' ')))
    # include folder name for alternative short forms
    if game_folder:
        _aliases.extend(generate_aliases(game_folder, split_words=False))
    # include lowercase version of app name in aliases
    if app_name:
        _aliases.append(app_name.lower())
    # include initialisms
    if len(game_parts) > 1:
        _aliases.append(''.join(p[0] for p in game_parts))
        _aliases.append(''.join(p[0] if p not in roman else p for p in game_parts))
        _aliases.append(''.join(roman.get(p, p[0]) for p in game_parts))
    # Attempt to address cases like "Battlefront" being shortened to "BF"
    if split_words:
        new_game_parts = []
        for word in game_parts:
            if len(word) >= 8 and word[3:-3].count('f') == 1:
                word_middle = word[3:-3]
                word_split = ' f'.join(word_middle.split('f'))
                word = word[0:3] + word_split + word[-3:]
                new_game_parts.extend(word.split())
            else:
                new_game_parts.append(word)

        if len(new_game_parts) > 1:
            _aliases.append(''.join(p[0] for p in new_game_parts))
            _aliases.append(''.join(p[0] if p not in roman else p for p in new_game_parts))
            _aliases.append(''.join(roman.get(p, p[0]) for p in new_game_parts))

    # return sorted uniques
    return sorted(set(_aliases))
